/////////////////////////////////////////////////////////////////////////////
// FPGA PACMAN TOP
//
// Version : 0.92
//
// Copyright(c) 2002-2004 Tatsuyuki Satoh , All rights reserved
//
// Important !
//
// This software is freeware for non-commercial use. 
// This software is provided "AS IS", without warranty of any kind.
// You can use this under your own risk. 
//
// note:
//
//  memory map when CFG_MODE==1
// +----------------+-------------------------------------------------------+
// |00000h-0ffffh.R |external memory (PROM/SRAM)                            |
// |00000h-0ffffh.W |external memory (SRAM)                                 |
// |04c00h-04cffh.RW|write through emem to waveform ROM                     |
// |04d00h-04dffh.RW|write through emem to CLUT ROM                         |
// |04e00h-04e1fh.RW|write through emem to PALETTE ROM                      |
// |04e40h-04e7fh.RW|write through emem to YPrPb table ROM for HDTV         |
// |04e80h-04ebfh.RW|write through emem to YUV table ROM for NTSC(reserved) |
// +---------------+--------------------------------------------------------+
//
/////////////////////////////////////////////////////////////////////////////

// design select and option
`include "src/switch.v"

/////////////////////////////////////////////////////////////////////////////
// module top
/////////////////////////////////////////////////////////////////////////////
module fpga_pac(
// Universal Video I/F
	VCLK,
	VGA_R,VGA_G,VGA_B,
	n_HSYNC,n_VSYNC,
//ROM interface
	RA,RD,RCS,ROE,RWE,
//Sound
	DAC_L,DAC_R,
//Contyroll switches
	HID,

`ifdef CFR_LOAD
// Configration ROM
  cfr_en, cfr_clk, cfr_di,
`endif

`ifdef MCPU_EXT_Z80
// external Z80
  ZA,
  ZDO,ZDI,
  ZMREQ_n,ZIORQ_n,ZM1_n,ZRD_n,ZWR_n, ZRFSH_n,
  ZCLK, ZINT_n, ZRESET_n, ZWAIT_n,
`endif

// 18.432MHz general CLOCK input
  CLK18M432
);
/////////////////////////////////////////////////////////////////////////////
// module I/O assign
/////////////////////////////////////////////////////////////////////////////
output DAC_L,DAC_R;

input CLK18M432; // 18.432MHz

// Universal Video I/F
input  VCLK;
output [4:0] VGA_R , VGA_G , VGA_B;
output n_HSYNC,n_VSYNC;

// code / pattern ROM interface
output [`ROM_MSB:0] RA;
inout [7:0] RD;
output RCS,ROE,RWE;

inout [7:0] HID;

`ifdef CFR_LOAD
// FPGA Configration ROM
output cfr_en;
inout cfr_clk;
input cfr_di;
`endif

`ifdef MCPU_EXT_Z80
// external Z80
input [15:0] ZA;
inout  [7:0] ZD;
input ZMREQ_n,ZIORQ_n,ZM1_n,ZRD_n,ZWR_n, ZRFSH_n;
output ZCLK, ZINT_n, ZRESET_n, ZWAIT_n;
`endif

/////////////////////////////////////////////////////////////////////////////
// basic timming signals
/////////////////////////////////////////////////////////////////////////////
wire [2:0] phase3x;
wire CLK12M288;
wire CLK6M144;
wire [8:0] hcnt;
wire [8:0] vcnt;

/////////////////////////////////////////////////////////////////////////////
// HumanInterface
/////////////////////////////////////////////////////////////////////////////
wire [7:0] hid_i , hid_o , hid_t;
wire [1:0] UP,LT,RT,DN,CO,ST;
wire SV;
wire user_rst_n;
wire [1:0] lamp;
wire coin_up;

`HID_MODULE HumanInterfaceDevice(
  .CLK(CLK18M432),
// Universal HID Bus
  .HID_I(hid_i),.HID_O(hid_o),.HID_T(hid_t),
// Internal Controll
  .UP(UP),.DN(DN),.LT(LT),.RT(RT),.B1(),.B2(),
  .COIN(CO),.START(ST),.SERVICE(SV),.RESET_N(user_rst_n),
// External Controll
  .LED(lamp)
);

/////////////////////////////////////////////////////////////////////////////
// Video encoder
/////////////////////////////////////////////////////////////////////////////

wire [1:0] blue;
wire [2:0] red,green;
wire hsync,vsync;

// video clock
wire video_clock;
//IBUFG video_clock_i(.I(VCLK),.O(video_clock));
assign video_clock = VCLK;

wire [4:0] r_out,g_out,b_out;
wire [4:0] r_tri,g_tri,b_tri;
wire hs_out , vs_out;
wire hs_tri , vs_tri;

/////////////////////////////////////////////////////////////////////////////
// On Screen Display
/////////////////////////////////////////////////////////////////////////////
`ifdef ENABLE_OSD
wire [7:0] rgb_osd = { red,green,blue };
wire [7:0] rgb_in;

pac_osd OnScreenDisplay(
// timming
  .pclk(CLK6M144),
  .hcnt(hcnt),
  .vcnt(vcnt),
// OSD source
  .lamp(lamp),
  .coin_up(coin_up),
// RGB signal
  .rgb8_in(rgb_osd),
  .rgb8_out(rgb_in)
);

`else
wire [7:0] rgb_in = { red,green,blue };
`endif

`VENC_MODULE pacman_video_encoder(
// clock input
	.pclk(CLK6M144),
	.dclk(CLK12M288),
	.vclk(video_clock),
// input
	.rgb8_in(rgb_in),
	.hs_in(hsync), .vs_in(vsync),
//output
	.r_out(r_out),   .r_tri(r_tri),
	.g_out(g_out),   .g_tri(g_tri),
	.b_out(b_out),   .b_tri(b_tri),
	.hs_out(hs_out), .hs_tri(hs_tri),
	.vs_out(vs_out), .vs_tri(vs_tri)
);

/////////////////////////////////////////////////////////////////////////////
// setup circuit
//
// upload all ROM data from CFG-ROM to external SRAM and internal ROM.
//   or
// upload internal ROM data from external PROM.
//
////////////////////////////////////////////////////////////////////////////
wire cfg_en;
wire [15:0] cfg_a;
wire [7:0] cfg_do;
wire cfg_dt,cfg_oe,cfg_we;

///////////////////////////////////////////////////////////
// upload external PROM to internal ROM
///////////////////////////////////////////////////////////
`ifdef PROM_LOAD

pac_prom setup_introm(
// timming signal
  .clk3x(CLK18M432) , .phase3x(phase3x),
// External PROM controll
  .CFG_EN(cfg_en),.CFG_A(cfg_a), .CFG_DO(cfg_do),
  .CFG_DT(cfg_dt),.CFG_OE(cfg_oe),.CFG_WE(cfg_we)
);
`endif // PROM_LOAD

///////////////////////////////////////////////////////////
// upload configrationROM to external SRAM and internal ROM
///////////////////////////////////////////////////////////
`ifdef CFR_LOAD

// Configration ROM controll pin
wire cfr_clk_o; // config-clock outout level
wire cfr_clk_t; // config-clock tristate controll

pac_prom setup_extram_introm(
// timming signal
  .clk3x(CLK18M432) , .phase3x(phase3x),
// configration ROM interface
  .CFR_EN(cfr_en), .CFR_DI(cfr_di),
  .CFR_CLK_I(cfr_clk), .CFR_CLK_O(cfr_clk_o), .CFR_CLK_T(cfr_clk_t),
// external SRAM & internal ROM controll
  .CFG_EN(cfg_en),.CFG_A(cfg_a), .CFG_DO(cfg_do),
  .CFG_DT(cfg_dt),.CFG_OE(cfg_oe),.CFG_WE(cfg_we)
);
assign cfr_clk = cfr_clk_t ? 1'bz : cfr_clk_o;

`endif // CFR_LOAD

/////////////////////////////////////////////////////////////////////////////
// internal memory 
//  - tilemap vram
//  - color vram
//  - work / object ram
//  - object ram2
//  - sound waveform ROM (write through 4c00-4cff from ext-mem in setup mode)
//  - color loocup rom   (write through 4d00-4dff from ext-mem in setup mode)
//  - palette rom        (write through 4e00-4e1f from ext-mem in setup mode)
/////////////////////////////////////////////////////////////////////////////
wire [11:0] AB;
wire [7:0] DB_I;
wire [7:0] DB_O;
wire tile_we , col_we , wram_we;
wire tile_cs , col_cs , wram_cs;

// object position ram
wire pram_we;
wire [7:0] pram_di,pram_do;

wire [7:0] lram_a;
wire [3:0] lram_di;
wire [3:0] lram_do;
wire lram_cs,lram_we;

wire [7:0] wave_a;
wire [3:0] wave_d;

wire [7:0] clut_a;
wire [3:0] clut_d;

wire [3:0] pal_a;
wire [7:0] pal_d;

wire cfg_wth = cfg_en & (cfg_oe | cfg_we); // read or write emem

`IMEM_MODULE Internalmemory(
  .CLK3X(CLK18M432),
// setup ROM data write bus
  .CFG_EN(cfg_en),.CFG_A(cfg_a),.CFG_DI(RD),.CFG_WE(cfg_wth),
// Tile/Code/Work VRAM
  .AB(AB),.DB_I(DB_O),.DB_O(DB_I),
  .TILE_CS(tile_cs),.TILE_WE(tile_we),
  .COL_CS(col_cs),.COL_WE(col_we),
  .WORK_CS(wram_cs),.WORK_WE(wram_we),
// Onject Position RAM
  .PWE(pram_we),.PDI(pram_do),.PDO(pram_di),
// Onject line RAM
  .LA(lram_a),.LDO(lram_di),.LDI(lram_do),.LCS(lram_cs),.LWE(lram_we),
// sound waveform ROM
  .WA(wave_a),.WDO(wave_d),
// color lookup table ROM
  .CLUTA(clut_a),.CLUT_DO(clut_d),
// palette table ROM (RGB / HDTV / NTSC)
  .PAL_A(pal_a),.PAL_SEL(2'b00),.PAL_DO(pal_d)
);

/////////////////////////////////////////////////////////////////////////////
// external memory interface
//  -CPU code ROM
//  -patterm ROM
/////////////////////////////////////////////////////////////////////////////

// external ROM signal
wire [15:0] rom_a;
wire [7:0]  rom_di;
wire [7:0]  rom_do;
wire rom_we,rom_t,rom_oe;

// read data
wire [15:0] ma;
wire [7:0] mrd;
wire moe;

// tile/obj pattern
wire [12:0] pa;
wire [7:0] prd;

pac_emem external_memory(
// Sync Clock
  .CLK3x(CLK18M432),
  .PHASE3x(phase3x),
  .HCNT(hcnt[1:0]),
// Static Memory I/F
  .MEM_A(rom_a),.MEM_DI(rom_di),.MEM_DO(rom_do),
  .MEM_WE(rom_we),.MEM_WO(rom_t),.MEM_OE(rom_oe),
// Main CPU
  .MROM_A(ma),.MROM_DO(mrd),.MROM_OE(moe),
// Pattern ROM
  .PROM_A(pa),.PROM_DO(prd)
);

/////////////////////////////////////////////////////////////////////////////
// sound encoder
/////////////////////////////////////////////////////////////////////////////
wire dac_out;
wire [9:0] wave_out;

// Sigma Delta
`SENC_MODULE sound_encoder(
  .RST_n(user_rst_n),
  .CLK(CLK18M432),
  .IN(wave_out),
  .OUT(dac_out)
);

/////////////////////////////////////////////////////////////////////////////
// Z80 core
/////////////////////////////////////////////////////////////////////////////

//----------- FZ80 -----------
`ifdef MCPU_FZ80C
wire [15:0] ZA;
wire [7:0] ZDO,ZDI;
wire ZMREQ_n,ZIORQ_n,ZM1_n,ZRD_n,ZWR_n, ZRFSH_n;
wire ZCLK, ZINT_n, ZRESET_n, ZWAIT_n;
fz80c Z80(
  .reset_n(ZRESET_n), 
  .clk(ZCLK),
  .A(ZA), .do(ZDO), .di(ZDI),
  .m1_n(ZM1_n), .mreq_n(ZMREQ_n), .iorq_n(ZIORQ_n), 
  .rd_n(ZRD_n), .wr_n(ZWR_n),
  .wait_n(ZWAIT_n), .rfsh_n(ZRFSH_n),
  .int_n(ZINT_n), .nmi_n(1'b1),.busrq_n(1'b1),
  .halt_n(), .busak_n()
);
`endif

//----------- TV80 -----------
`ifdef MCPU_TV80
wire [15:0] ZA;
wire [7:0] ZDO,ZDI;
wire ZMREQ_n,ZIORQ_n,ZM1_n,ZRD_n,ZWR_n, ZRFSH_n;
wire ZCLK, ZINT_n, ZRESET_n, ZWAIT_n;
tv80c Z80(
  .reset_n(ZRESET_n),
  .clk(ZCLK),
  .A(ZA), .do(ZDO), .di(ZDI),
  .m1_n(ZM1_n), .mreq_n(ZMREQ_n), .iorq_n(ZIORQ_n), 
  .rd_n(ZRD_n), .wr_n(ZWR_n),
  .wait_n(ZWAIT_n), .rfsh_n(ZRFSH_n),
  .int_n(ZINT_n), .nmi_n(1'b1),.busrq_n(1'b1),
  .halt_n(), .busak_n()
);
`endif

//----------- extend Z80 -----------
`ifdef MCPU_EXT_Z80
// external Z80
assign ZD = ZRD_n ? 8'bzzzzzzzz : ZDO;
assign ZDI = ZD;
`endif

/////////////////////////////////////////////////////////////////////////////
// pacman main pcb
/////////////////////////////////////////////////////////////////////////////

wire pacman_reset_n = user_rst_n & ~cfg_en;

pacman #(`DSW1,`DSW2) pacman(
  .reset_n(pacman_reset_n),
// 18.432MHz general CLOCK input
  .CLK18M432(CLK18M432),
  .PHASE3X(phase3x),
  .CLK12M288(CLK12M288),
  .CLK6M144(CLK6M144),
//
  .hcnt(hcnt),.vcnt(vcnt),
  .hsync(hsync),.vsync(vsync),
  .blue(blue),.red(red),.green(green),

//Sound
  .wave_out(wave_out),
//Z80
  .ZRESET_n(ZRESET_n),
  .ZCLK(ZCLK),
  .ZA(ZA), .ZDO(ZDO), .ZDI(ZDI),
  .ZM1_n(ZM1_n), .ZMREQ_n(ZMREQ_n), .ZIORQ_n(ZIORQ_n), 
  .ZRD_n(ZRD_n), .ZWR_n(ZWR_n),
  .ZWAIT_n(ZWAIT_n), .ZRFSH_n(ZRFSH_n),
  .ZINT_n(ZINT_n),
// CPU code rom
  .ma(ma),.md_i(mrd),.moe(moe),
// Pattern ROM
  .pa(pa),.pd_i(prd),

// tile,color,obj2/work,
  .AB(AB),
  .DB_I(DB_I),
  .DB_O(DB_O),
  .tile_cs(tile_cs), .tile_we(tile_we),
  .col_cs(col_cs),   .col_we(col_we),
  .wram_cs(wram_cs), .wram_we(wram_we),
// obj2 ram
  .pram_di(pram_di), .pram_do(pram_do),
  .pram_we(pram_we),
// object line ram
  .lram_a(lram_a),
  .lram_di(lram_di), .lram_do(lram_do),
  .lram_cs(lram_cs), .lram_we(lram_we),
// sound waveform rom
  .wave_a(wave_a),
  .wave_d(wave_d),
// color lookup table rom
  .clut_a(clut_a),
  .clut_d(clut_d),
// palette rom
  .pal_a(pal_a),
  .pal_d(pal_d),
//Contyroll switches
  .UP(UP),.LT(LT),.RT(RT),.DN(DN),.CO(CO),.ST(ST),.SV(SV),
  .LAMP(lamp),.COINUP(coin_up)
);

/////////////////////////////////////////////////////////////////////////////
// ROM I/F
/////////////////////////////////////////////////////////////////////////////
assign RA  = cfg_en ? {`ROM_BANK,cfg_a} : {`ROM_BANK,rom_a};
assign RCS = 1'b0;
assign RWE = cfg_en ? ~cfg_we : 1'b1;
assign ROE = cfg_en ? ~cfg_oe : 1'b0;
assign rom_di = RD;
assign RD = cfg_dt ? 8'bzzzzzzzz : cfg_do;

/////////////////////////////////////////////////////////////////////////////
// Universal Video Output Interface
//
// HS,HV = direct connect with 47ohm DUMP/safe Register
/////////////////////////////////////////////////////////////////////////////

assign VGA_R[0] = r_tri[4] ? 1'bz : r_out[4];
assign VGA_R[1] = r_tri[3] ? 1'bz : r_out[3];
assign VGA_R[2] = r_tri[2] ? 1'bz : r_out[2];
assign VGA_R[3] = r_tri[1] ? 1'bz : r_out[1];
assign VGA_R[4] = r_tri[0] ? 1'bz : r_out[0];

assign VGA_G[0] = g_tri[4] ? 1'bz : g_out[4];
assign VGA_G[1] = g_tri[3] ? 1'bz : g_out[3];
assign VGA_G[2] = g_tri[2] ? 1'bz : g_out[2];
assign VGA_G[3] = g_tri[1] ? 1'bz : g_out[1];
assign VGA_G[4] = g_tri[0] ? 1'bz : g_out[0];

assign VGA_B[0] = b_tri[4] ? 1'bz : b_out[4];
assign VGA_B[1] = b_tri[3] ? 1'bz : b_out[3];
assign VGA_B[2] = b_tri[2] ? 1'bz : b_out[2];
assign VGA_B[3] = b_tri[1] ? 1'bz : b_out[1];
assign VGA_B[4] = b_tri[0] ? 1'bz : b_out[0];

assign n_HSYNC = hs_tri ? 1'bz : hs_out;
assign n_VSYNC = vs_tri ? 1'bz : vs_out;

/////////////////////////////////////////////////////////////////////////////
// SOUND DAC interface
/////////////////////////////////////////////////////////////////////////////

assign DAC_L = dac_out;
assign DAC_R = dac_out;

/////////////////////////////////////
// Pin Assign
/////////////////////////////////////
assign HID[0] = hid_t[0] ? 1'bz : hid_o[0];
assign HID[1] = hid_t[1] ? 1'bz : hid_o[1];
assign HID[2] = hid_t[2] ? 1'bz : hid_o[2];
assign HID[3] = hid_t[3] ? 1'bz : hid_o[3];
assign HID[4] = hid_t[4] ? 1'bz : hid_o[4];
assign HID[5] = hid_t[5] ? 1'bz : hid_o[5];
assign HID[6] = hid_t[6] ? 1'bz : hid_o[6];
assign HID[7] = hid_t[7] ? 1'bz : hid_o[7];
assign hid_i = HID;

endmodule
