/////////////////////////////////////////////////////////////////////////////
//
// Interleaved External Memory Interface for PACMAN
//
// Version : 0.92
//
// Copyright(c) 2004 Tatsuyuki Satoh , All rights reserved
//
// Important !
//
// This software is freeware for non-commercial use. 
// This software is provided "AS IS", without warranty of any kind.
// You can use this under your own risk. 
//
// Histry:
//   2004. 9.12 1st Release
//
/////////////////////////////////////////////////////////////////////////////

/****************************************************************************
  Memory Map

 +---------+-----+-----------------------------------------+
 |ROM Area | LOC | Contents                                |
 +---------+-----+-----------------------------------------+
 |0000-0fff| 6E  |MainCPU ROM 0000-0fff                    |
 |1000-1fff| 6F  |MainCPU ROM 1000-1fff                    |
 |2000-2fff| 6H  |MainCPU ROM 2000-2fff                    |
 |3000-3fff| 6J  |MainCPU ROM 3000-3ffff                   |
 +---------+-----+-----------------------------------------+
 |4000-40ff| --  | Original BootROM       (CPU 0000-00ff)*1|
 |4100-41ff| 1M  | sound waveform ROM     (CPU 0100-01ff)*1|
 |4200-42ff| 4A  | Color LookUp Table ROM (CPU 0200-02ff)*1|
 |4300-431f| 7F  | palette ROM            (CPU 0200-02ff)*1|
 +---------+-----+-----------------------------------------+
 |6000-6fff| 5E  |Tilemap Patterm ROM                      |
 |7000-7fff| 5F  |Object Pattern ROM                       |
 +---------+-----+-----------------------------------------+
 |8000-bfff| --  |MainCPU ROM 8000-bfff                    |
 +---------+-----+-----------------------------------------+
 *1: These are mapped to CPU 0000-031f at power up.
     BootROM code transfer some ROM to FPGA interenal RAM,
     and reset wirh normal memory map.

****************************************************************************/

///////////////////////////////////////////////////////////////////////////

module pac_emem(
// Sync Clock
  CLK3x, PHASE3x, HCNT,
// Static Memory I/F
  MEM_A,MEM_DI,MEM_DO,MEM_WE,MEM_WO,MEM_OE,
// Main CPU
  MROM_A,MROM_DO,MROM_OE,
// Pattern ROM
  PROM_A,PROM_DO
);

/////////////////////////////////////////////////////////////////////////////
// module I/O assign
/////////////////////////////////////////////////////////////////////////////
input CLK3x;         // x3 pixel clock (18.432MHz)
input [2:0] PHASE3x; // x3 clock phase
input [1:0] HCNT;

output [15:0] MEM_A;
input  [7:0] MEM_DI;
output [7:0] MEM_DO;
output MEM_WE;
output MEM_WO;
output MEM_OE;

// Main CPU
input  [15:0] MROM_A;
output [7:0] MROM_DO;
input MROM_OE;

// Pattern ROM
input  [12:0] PROM_A;
output [7:0] PROM_DO;

/////////////////////////////////////////////////////////////////////////////

wire [15:0] cpu_a  = { 1'b0   , MROM_A };  //0000-3fff / 8000-bfff
wire [15:0] pat_a  = { 3'b011 , PROM_A };  //6000-7fff

//
// interleave ROM
//

reg [15:0] rom_a;
reg [7:0] mrom_r , prom_r;

always @(posedge CLK3x)
begin
  if(PHASE3x[1])
  begin
    case(HCNT[0])
    1'b0:begin rom_a <= cpu_a; prom_r <= MEM_DI; end
    1'b1:begin rom_a <= pat_a; mrom_r <= MEM_DI; end
    endcase
  end
end

assign MEM_A  = rom_a;
assign MEM_DO = 8'h00;
assign MEM_WO = 1'b0;
assign MEM_WE = 1'b0;
assign MEM_OE = 1'b1;

assign PROM_DO = prom_r;
assign MROM_DO = MROM_OE ? mrom_r : 8'h00;

endmodule
