/////////////////////////////////////////////////////////////////////////////
// FPGA PACMAN TOP
//
// Version : 0.92
//
// Copyright(c) 2002-2004 Tatsuyuki Satoh , All rights reserved
//
// Important !
//
// This software is freeware for non-commercial use. 
// This software is provided "AS IS", without warranty of any kind.
// You can use this under your own risk. 
//
// Histry:
//    2004. 9. 9 export memory and external circuit
//
/////////////////////////////////////////////////////////////////////////////

module pacman(
  reset_n,
// 18.432MHz general CLOCK input
  CLK18M432,
// Timming Signal
  PHASE3X,
  CLK12M288,
  CLK6M144,
  hcnt,vcnt,
// Video output
  hsync,vsync,
  red,green,blue,
//Sound output
  wave_out,
//Z80 CPU
  ZRESET_n,
  ZCLK,
  ZA,
  ZDO,
  ZDI,
  ZM1_n,
  ZMREQ_n, 
  ZIORQ_n,
  ZRD_n,
  ZWR_n,
  ZWAIT_n,
  ZRFSH_n,
  ZINT_n,
// CPU code rom
  ma,md_i,moe,
// Pattern ROM
  pa,pd_i,
// tile,color,obj2/work vram
  AB, DB_I, DB_O,
  tile_cs,tile_we,
  col_cs ,col_we,
  wram_cs ,wram_we,
// obj2 video ram
  pram_di,pram_do,pram_we,
// object line ram
  lram_a,lram_di,lram_do,lram_cs,lram_we,
// sound waveform rom
  wave_a,wave_d,
// color lookup table rom
  clut_a,clut_d,
// palette rom
  pal_a,pal_d,

//Contyroll switches
  UP,LT,RT,DN,CO,ST,SV,
  LAMP,COINUP
);
/////////////////////////////////////////////////////////////////////////////
// module I/O assign
/////////////////////////////////////////////////////////////////////////////

parameter DSW1 = 8'b00000000;
parameter DSW2 = 8'b00000000;

input reset_n;

input CLK18M432; // 18.432MHz

// Timming signal
output [2:0] PHASE3X;
output CLK12M288;
output CLK6M144;
output [8:0] hcnt;
output [8:0] vcnt;

// Video Output
output hsync,vsync;
output [2:0] red;
output [2:0] green;
output [1:0] blue;

// Sound Output
output [9:0] wave_out;

//Z80 CPU
output ZRESET_n;
output ZCLK;
output ZWAIT_n;
output ZINT_n;
output [7:0] ZDI;
input [15:0] ZA;
input  [7:0] ZDO;
input ZM1_n;
input ZMREQ_n;
input ZIORQ_n;
input ZRD_n;
input ZWR_n;
input ZRFSH_n;

// CPU code rom
output [15:0] ma;
input [7:0] md_i;
output moe;

// Pattern ROM
output [12:0] pa; //patterm ROM address
input  [7:0] pd_i;  //pattern ROM data

// ram
output [11:0] AB;
input [7:0] DB_I;
output [7:0] DB_O;
output tile_we , col_we , wram_we;
output tile_cs , col_cs , wram_cs;

output pram_we;
input  [7:0] pram_di;
output [7:0] pram_do;

output [7:0] lram_a;
input  [3:0] lram_di;
output [3:0] lram_do;
output lram_cs,lram_we;

output [7:0] wave_a;
input  [3:0] wave_d;

output [7:0] clut_a;
input  [3:0] clut_d;

output [3:0] pal_a;
input  [7:0] pal_d;

input [1:0] UP,LT,RT,DN,CO,ST;
input SV;

output [1:0] LAMP;
output COINUP;

/////////////////////////////////////////////////////////////////////////////
// clock generator
/////////////////////////////////////////////////////////////////////////////

pac_clk BASE_CLOCK(
	.CLKIN(CLK18M432),
	.PHASE3X(PHASE3X),
	.CLK_2_3(CLK12M288),
	.CLK_1_6(CLK6M144)
);

/////////////////////////////////////////////////////////////////////////////
// address decoder
/////////////////////////////////////////////////////////////////////////////

wire flip;
wire cs_in0,cs_in1,cs_dsw1,cs_dsw2;
wire cs_wr0,cs_wr1;
wire snd_on;
wire cs_xyram , cs_vram;
pac_adec pacman_adec(
	.RST(~ZRESET_n),
// inputs
	.PCLK(CLK6M144),
	.A(ZA),
	.n_MREQ(ZMREQ_n),
	.D0I(ZDO[0]),
	.n_RD(ZRD_n),
	.n_WR(ZWR_n),
	.n_RFSH(ZRFSH_n),
// decode output
	.ROM(moe),
	.RAM(cs_vram),
	.IN0(cs_in0),
	.IN1(cs_in1),
	.DSW1(cs_dsw1),
	.DSW2(cs_dsw2),
	.WR0(cs_wr0),
	.WR1(cs_wr1),	// 5050-505f.W sound frequency   / volume
	.WR2(cs_xyram),	// 5060-506f.W object X,Y RAM
	.WDT(),	// 50c0-50ff.W Watchdog timer reset
// latched register
	.IRQ(irqen),	// 5000.W      IRQ enable (clear)
	.SOUND_ON(snd_on),// 5001.W      sound enable
	.FLIP(flip),	// 5003.W      flip
	.LAMP1P(LAMP[0]), // 5004.W      1P start lamp
	.LAMP2P(LAMP[1]), // 5005.W      2P start lamp
	.COINLK(),	// 5006.W      coin lockout
	.COINUP(COINUP)	// 5007.W      coin countup
);

assign ma = ZA;

/////////////////////////////////////////////////////////////////////////////
// controller IN / DIP switch I/F
/////////////////////////////////////////////////////////////////////////////

// wire usr_rst_n;

//wire L1 = PSW[3];//L
//wire R1 = PSW[1];//R
//wire U1 = PSW[0];
//wire D1 = PSW[2];//D

/////////////////////////////////////////////////////////////////////////////
// inport
/////////////////////////////////////////////////////////////////////////////

wire [7:0] inp_do;
pac_inp #(DSW1,DSW2) pacman_inport(
// inputs
	.LEFT1(LT[0]),
	.RIGHT1(RT[0]),
	.UP1(UP[0]),
	.DOWN1(DN[0]),
	.LEFT2(LT[1]),
	.RIGHT2(RT[1]),
	.UP2(UP[1]),
	.DOWN2(DN[1]),
	.START1(ST[0]),
	.START2(ST[1]),
	.COIN1(CO[0]),
	.COIN2(CO[1]),
	.CREDIT(SV),
	.TEST(1'b1),
	.TABLE(1'b1),
//enables
	.IN0_EN(cs_in0),
	.IN1_EN(cs_in1),
	.DSW1_EN(cs_dsw1),
	.DSW2_EN(cs_dsw2),
//output
	.DO(inp_do)	
);

/////////////////////////////////////////////////////////////////////////////
// VIDEO hardware interface
/////////////////////////////////////////////////////////////////////////////

//wire [8:0] hcnt;
//wire hsync,vsync;
wire [7:0] vdo;
wire vblank;

pac_vid vh(
	.A(ZA[11:0]),
	.csvram(cs_vram),
	.csxyram(cs_xyram),
	.n_rd(ZRD_n),
	.di(ZDO),
	.do(vdo),
	.pclk(CLK6M144),
	.flip(flip),
	.vb(vblank),
	.hc(hcnt),
	.vc(vcnt),
	.hs(hsync),
	.vs(vsync),
//	.r(red),
//	.g(green),
//	.b(blue),
	.CA(pa),
	.CD(pd_i),
// CPU & VIDEO memory bus , & sound timming address
	.AB(AB), 
// VIDEO memory bus & controll
	.DB_I(DB_I),
	.TRAM_WE(tile_we),
	.TRAM_CS(tile_cs),
	.CRAM_WE(col_we),
	.CRAM_CS(col_cs),
	.WRAM_WE(wram_we),
	.WRAM_CS(wram_cs),
// line ram
	.LRAM_A(lram_a),
	.LRAM_DI(lram_di),
	.LRAM_DO(lram_do),
	.LRAM_CS(lram_cs),
	.LRAM_WE(lram_we),
// object position ram
	.PRAM_DI(pram_di),
	.PRAM_DO(pram_do),
	.PRAM_WE(pram_we),
// CLUT ROM
	.CLUT_A(clut_a),
	.CLUT_DI(clut_d),
// PaletteROM
	.PAL_A(pal_a)
);

/////////////////////////////////////////////////////////////////////////////
// CPU output signal
/////////////////////////////////////////////////////////////////////////////

//BUFG ZCLK(.I(~hcnt[0]),.O(ZCLK)); // 3MHz
assign ZCLK = ~hcnt[0];
assign ZWAIT_n = ~(cs_vram & hcnt[1] & ~ZRD_n); // wait controll
assign ZRESET_n = reset_n;

/////////////////////////////////////////////////////////////////////////////
// IRQ & IRQ vector
/////////////////////////////////////////////////////////////////////////////

wire [7:0] irq_do;
pac_irq pacman_irqvector(
//inputs
	.n_M1(ZM1_n),
	.n_IORQ(ZIORQ_n),
	.n_WR(ZWR_n),
	.IRQEN(irqen),
	.IRQTRG(vblank),
	.DI(ZDO),
//outputs
	.n_IRQ(ZINT_n),
	.DO(irq_do),
	.DE()
);

/////////////////////////////////////////////////////////////////////////////
// sound
/////////////////////////////////////////////////////////////////////////////

pac_snd pacman_sound(
	.AB(ZA[3:0]),
//	.AB(AB[3:0]),
	.DI(ZDO[3:0]),
	.WR0(cs_wr0),
	.WR1(cs_wr1),
	.PCLK(CLK6M144),
	.SOUND_ON(snd_on),
	.HC(hcnt[5:0]),
//
	.DAO(wave_out),
	.WA(wave_a),
	.WDI(wave_d)
);

/////////////////////////////////////////////////////////////////////////////
// R,G,B output signal
/////////////////////////////////////////////////////////////////////////////

assign blue    = pal_d[7:6];
assign green   = pal_d[5:3];
assign red     = pal_d[2:0];

assign DB_O = ZDO;

/////////////////////////////////////////////////////////////////////////////
// data bus multiplexer
/////////////////////////////////////////////////////////////////////////////

assign ZDI = moe ? md_i : vdo | inp_do | irq_do;

endmodule
